import array


def assert_raises(exc_type, func, contains=None):
    try:
        func()
    except Exception as e:
        assert isinstance(e, exc_type), f'expected {exc_type}, got {type(e)}: {e}'
        if contains is not None:
            assert contains in str(e), f"expected '{contains}' in error: {e}"
    else:
        assert False, f'expected {exc_type} to be raised'


# === Module API ===
for tc in 'bBuhHiIlLqQfd':
    assert tc in array.typecodes
assert array.ArrayType is array.array

# === Constructor and attributes ===
a = array.array('i', [1, 2, 3])
assert a.typecode == 'i'
assert a.itemsize == 4
assert len(a) == 3
assert repr(a) == "array('i', [1, 2, 3])"

assert array.array('i').tolist() == []
assert array.array('b', b'\x01\xff').tolist() == [1, -1]
assert array.array('u', 'ab').tolist() == ['a', 'b']
assert repr(array.array('u', 'ab')) == "array('u', 'ab')"

# === Basic mutation APIs ===
a.append(4)
assert a.tolist() == [1, 2, 3, 4]

a.extend([5, 6])
assert a.tolist() == [1, 2, 3, 4, 5, 6]

a.extend(array.array('i', [7, 8]))
assert a.tolist() == [1, 2, 3, 4, 5, 6, 7, 8]
assert_raises(TypeError, lambda: a.extend(array.array('h', [9])), 'same kind')

a.insert(0, -1)
a.insert(999, 9)
a.insert(-999, -2)
assert a.tolist() == [-2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9]

assert a.pop() == 9
assert a.pop(0) == -2
assert_raises(IndexError, lambda: a.pop(999), 'range')

a.remove(-1)
assert -1 not in a
assert_raises(ValueError, lambda: a.remove(9999), 'not in array')

# === Indexing and slicing ===
b = array.array('i', [0, 1, 2, 3, 4, 5])
assert b[0] == 0
assert b[-1] == 5
assert b[1:5].tolist() == [1, 2, 3, 4]
assert b[::-1].tolist() == [5, 4, 3, 2, 1, 0]

b[1] = 100
assert b.tolist() == [0, 100, 2, 3, 4, 5]

b[2:4] = array.array('i', [20, 30, 40])
assert b.tolist() == [0, 100, 20, 30, 40, 4, 5]

b[::2] = array.array('i', [9, 8, 7, 6])
assert b.tolist() == [9, 100, 8, 30, 7, 4, 6]

# explicit extended-slice mismatch on assignment target
def _assign_extended_slice_mismatch():
    x = array.array('i', [0, 1, 2, 3])
    x[::2] = array.array('i', [5])


def _assign_list_to_slice():
    x = array.array('i', [0, 1])
    x[:1] = [9]


assert_raises(ValueError, _assign_extended_slice_mismatch, 'extended slice')
assert_raises(TypeError, _assign_list_to_slice, 'array slice')

c = array.array('i', [10, 11, 12, 13, 14, 15])
del c[1]
assert c.tolist() == [10, 12, 13, 14, 15]
del c[1:4]
assert c.tolist() == [10, 15]

d = array.array('i', [0, 1, 2, 3, 4, 5, 6])
del d[::2]
assert d.tolist() == [1, 3, 5]

# === Query APIs ===
e = array.array('i', [1, 2, 2, 3])
assert 2 in e
assert 9 not in e
assert e.index(2) == 1
assert_raises(ValueError, lambda: e.index(9), 'not in array')
assert e.count(2) == 2

e.reverse()
assert e.tolist() == [3, 2, 2, 1]

# === Iteration ===
assert list(e.__iter__()) == [3, 2, 2, 1]
if hasattr(e, '__reversed__'):
    assert list(e.__reversed__()) == [1, 2, 2, 3]
else:
    assert list(reversed(e)) == [1, 2, 2, 3]

# === Conversion APIs ===
f = array.array('h', [0x1234, -2])
orig = f.tobytes()
assert isinstance(orig, bytes)

f2 = array.array('h')
f2.frombytes(orig)
assert f2.tolist() == [0x1234, -2]

f3 = array.array('h')
f3.fromlist([7, 8])
assert f3.tolist() == [7, 8]
assert_raises(TypeError, lambda: f3.fromlist((1, 2)), 'list')

assert f3.tolist() == [7, 8]

addr, length = f3.buffer_info()
assert isinstance(addr, int)
assert isinstance(length, int)
assert length == 2

# byteswap should be involutive
f.byteswap()
f.byteswap()
assert f.tobytes() == orig

# === Operators and comparisons ===
g = array.array('i', [1, 2])
h = array.array('i', [3])

assert (g + h).tolist() == [1, 2, 3]
assert_raises(TypeError, lambda: g + array.array('h', [3]), 'bad argument type')

x = array.array('i', [1])
x += array.array('i', [2, 3])
assert x.tolist() == [1, 2, 3]
def _iadd_non_array():
    x = array.array('i', [1])
    x += [2]


assert_raises(TypeError, _iadd_non_array, 'array')

assert (array.array('i', [4, 5]) * 2).tolist() == [4, 5, 4, 5]
assert (2 * array.array('i', [4, 5])).tolist() == [4, 5, 4, 5]

y = array.array('i', [9])
y *= 3
assert y.tolist() == [9, 9, 9]

assert array.array('i', [1, 2]) == array.array('h', [1, 2])
assert array.array('i', [1, 2]) != array.array('i', [1, 3])
assert array.array('i', [1, 2]) < array.array('i', [1, 3])
assert array.array('i', [1, 2]) <= array.array('i', [1, 2])
assert array.array('i', [2]) > array.array('i', [1, 100])
assert array.array('i', [2]) >= array.array('i', [2])

# === Roundtrip all typecodes ===
roundtrip_samples = {
    'b': [-2, 0, 2],
    'B': [0, 200, 255],
    'u': ['a', 'Z', '0'],
    'h': [-123, 0, 123],
    'H': [0, 123, 65535],
    'i': [-123456, 0, 123456],
    'I': [0, 123456, 2**32 - 1],
    'l': [-123456789, 0, 123456789],
    'L': [0, 123456789, 2**63],
    'q': [-(2**63), -1, 2**63 - 1],
    'Q': [0, 123456789, 2**64 - 1],
    'f': [0.5, -1.5, 10.0],
    'd': [0.5, -1.5, 10.0],
}

for tc, values in roundtrip_samples.items():
    arr = array.array(tc, values)
    blob = arr.tobytes()
    out = array.array(tc)
    out.frombytes(blob)
    assert out.tolist() == arr.tolist(), f'roundtrip mismatch for {tc}'

# === Error paths ===
assert_raises(TypeError, lambda: array.array(1), 'unicode character')
assert_raises(TypeError, lambda: array.array('zz'), 'length 2')
assert_raises(ValueError, lambda: array.array('?'), 'bad typecode')
assert_raises(TypeError, lambda: array.array('i', 'abc'), 'cannot use a str')
assert_raises(ValueError, lambda: array.array('i', b'abc'), 'item size')

assert_raises(OverflowError, lambda: array.array('B').append(-1), None)
assert_raises(OverflowError, lambda: array.array('B').append(256), None)
assert_raises(OverflowError, lambda: array.array('b').append(128), None)

assert_raises(TypeError, lambda: array.array('i').frombytes('abc'), 'bytes-like')
assert_raises(ValueError, lambda: array.array('i').frombytes(b'123'), 'item size')

empty = array.array('i')
assert_raises(IndexError, lambda: empty.pop(), 'empty')
assert_raises(IndexError, lambda: empty[0], 'range')
def _setitem_empty():
    x = array.array('i')
    x[0] = 1


def _delitem_empty():
    x = array.array('i')
    del x[0]


assert_raises(IndexError, _setitem_empty, 'range')
assert_raises(IndexError, _delitem_empty, 'range')



# Tests that values of different types are returned correctly
# Also tests identity operators with singletons

# === Boolean values ===
assert repr(False) == 'False', 'False repr'
assert repr(True) == 'True', 'True repr'

# === None value ===
assert repr(None) == 'None', 'None repr'

# === Ellipsis value ===
assert repr(...) == 'Ellipsis', 'Ellipsis repr'

# === Ellipsis identity ===
assert (... is ...) == True, 'ellipsis is ellipsis'
assert (None is ...) == False, 'none is not ellipsis'

# === Type checks against None ===
assert (False is None) == False, 'False is not None'
assert (True is None) == False, 'True is not None'
assert (None is None) == True, 'None is None'
assert (42 is None) == False, 'int is not None'
assert (3.14 is None) == False, 'float is not None'
assert ([1, 2] is None) == False, 'list is not None'
assert ('hello' is None) == False, 'str is not None'
assert ((1, 2) is None) == False, 'tuple is not None'

# === Type checks against Ellipsis ===
assert (False is ...) == False, 'False is not Ellipsis'
assert (True is ...) == False, 'True is not Ellipsis'
assert (None is ...) == False, 'None is not Ellipsis'
assert (42 is ...) == False, 'int is not Ellipsis'
assert (3.14 is ...) == False, 'float is not Ellipsis'
assert ([1, 2] is ...) == False, 'list is not Ellipsis'
assert ('hello' is ...) == False, 'str is not Ellipsis'
assert ((1, 2) is ...) == False, 'tuple is not Ellipsis'



import pytest
from inline_snapshot import snapshot

import ouros


def test_type_check_no_errors():
    """Type checking code with no errors returns None."""
    m = ouros.Sandbox('x = 1')
    assert m.type_check() is None


def test_type_check_with_errors():
    """Type checking code with type errors raises SandboxTypingError."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    assert str(exc_info.value) == snapshot("""\
error[unsupported-operator]: Unsupported `+` operation
 --> main.py:1:1
  |
1 | "hello" + 1
  | -------^^^-
  | |         |
  | |         Has type `Literal[1]`
  | Has type `Literal["hello"]`
  |
info: rule `unsupported-operator` is enabled by default

""")


def test_type_check_function_return_type():
    """Type checking detects mismatched return types."""
    code = """
def foo() -> int:
    return "not an int"
"""
    m = ouros.Sandbox(code)
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    assert str(exc_info.value) == snapshot("""\
error[invalid-return-type]: Return type does not match returned value
 --> main.py:2:14
  |
2 | def foo() -> int:
  |              --- Expected `int` because of return type
3 |     return "not an int"
  |            ^^^^^^^^^^^^ expected `int`, found `Literal["not an int"]`
  |
info: rule `invalid-return-type` is enabled by default

""")


def test_type_check_undefined_variable():
    """Type checking detects undefined variables."""
    m = ouros.Sandbox('print(undefined_var)')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    assert str(exc_info.value) == snapshot("""\
error[unresolved-reference]: Name `undefined_var` used when not defined
 --> main.py:1:7
  |
1 | print(undefined_var)
  |       ^^^^^^^^^^^^^
  |
info: rule `unresolved-reference` is enabled by default

""")


def test_type_check_valid_function():
    """Type checking valid function returns None."""
    code = """
def add(a: int, b: int) -> int:
    return a + b

add(1, 2)
"""
    m = ouros.Sandbox(code)
    assert m.type_check() is None


def test_type_check_with_prefix_code():
    """Type checking with prefix code for input declarations."""
    m = ouros.Sandbox('result = x + 1')
    # Without prefix, x is undefined
    with pytest.raises(ouros.SandboxTypingError):
        m.type_check()
    # With prefix declaring x as a variable, it should pass
    assert m.type_check(prefix_code='x = 0') is None


def test_type_check_display_invalid_format():
    """Invalid format string on display() raises ValueError."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    with pytest.raises(ValueError) as val_exc:
        exc_info.value.display('invalid_format')  # pyright: ignore[reportArgumentType]
    assert str(val_exc.value) == snapshot('Unknown format: invalid_format')


def test_type_check_display_concise_format():
    """Type checking with concise format via display()."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    assert exc_info.value.display('concise') == snapshot(
        'main.py:1:1: error[unsupported-operator] Operator `+` is not supported between objects of type `Literal["hello"]` and `Literal[1]`\n'
    )


# === SandboxTypingError tests ===


def test_ouros_typing_error_is_ouros_error_subclass():
    """SandboxTypingError is a subclass of SandboxError."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    error = exc_info.value
    assert isinstance(error, ouros.SandboxError)
    assert isinstance(error, Exception)


def test_ouros_typing_error_repr():
    """SandboxTypingError has proper repr with truncation."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    # repr truncates at 50 chars
    assert repr(exc_info.value) == snapshot("""\
SandboxTypingError(error[unsupported-operator]: Unsupported `+` operation
 --> main.py:1:1
  |
1 | "hello" + 1
  | -------^^^-
  | |         |
  | |         Has type `Literal[1]`
  | Has type `Literal["hello"]`
  |
info: rule `unsupported-operator` is enabled by default

)\
""")


def test_ouros_typing_error_caught_as_ouros_error():
    """SandboxTypingError can be caught as SandboxError."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxError):
        m.type_check()


def test_ouros_typing_error_display_default():
    """SandboxTypingError display() defaults to full format."""
    m = ouros.Sandbox('"hello" + 1')
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        m.type_check()
    # Default display should match str()
    assert exc_info.value.display() == str(exc_info.value)


# === Constructor type_check parameter tests ===


def test_constructor_type_check_default_false():
    """Type checking is disabled by default in constructor."""
    # This should NOT raise during construction (type_check=False is default)
    m = ouros.Sandbox('"hello" + 1')
    # But we can still call type_check() manually later
    with pytest.raises(ouros.SandboxTypingError):
        m.type_check()


def test_constructor_type_check_explicit_true():
    """Explicit type_check=True raises on type errors."""
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        ouros.Sandbox('"hello" + 1', type_check=True)
    assert str(exc_info.value) == snapshot("""\
error[unsupported-operator]: Unsupported `+` operation
 --> main.py:1:1
  |
1 | "hello" + 1
  | -------^^^-
  | |         |
  | |         Has type `Literal[1]`
  | Has type `Literal["hello"]`
  |
info: rule `unsupported-operator` is enabled by default

""")


def test_constructor_type_check_explicit_false():
    """Explicit type_check=False skips type checking during construction."""
    # This should NOT raise during construction
    m = ouros.Sandbox('"hello" + 1', type_check=False)
    # But we can still call type_check() manually later
    with pytest.raises(ouros.SandboxTypingError):
        m.type_check()


def test_constructor_default_allows_run_with_inputs():
    """Default (type_check=False) allows running code that would fail type checking."""
    # Code with undefined variable - type checking would fail
    m = ouros.Sandbox('x + 1', inputs=['x'])
    # But runtime works fine with the input provided
    result = m.run(inputs={'x': 5})
    assert result == 6


def test_constructor_type_check_stubs():
    """type_check_stubs provides declarations for type checking."""
    # Without prefix, this would fail type checking (x is undefined)
    # Use assignment to define x, not just type annotation
    m = ouros.Sandbox('result = x + 1', type_check=True, type_check_stubs='x = 0')
    # Should construct successfully because prefix declares x
    assert m is not None


def test_constructor_type_check_stubs_with_external_function():
    """type_check_stubs can declare external function signatures."""
    # Define fetch as a function that takes a string and returns a string
    prefix = """
def fetch(url: str) -> str:
    return ''
"""
    m = ouros.Sandbox(
        'result = fetch("https://example.com")',
        external_functions=['fetch'],
        type_check=True,
        type_check_stubs=prefix,
    )
    assert m is not None


def test_constructor_type_check_stubs_invalid():
    """type_check_stubs with wrong types still catches errors."""
    # Prefix defines x as str, but code tries to use it with int addition
    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        ouros.Sandbox(
            'result: int = x + 1',
            type_check=True,
            type_check_stubs='x = "hello"',
        )
    # Should fail because str + int is invalid
    assert str(exc_info.value) == snapshot("""\
error[unsupported-operator]: Unsupported `+` operation
 --> main.py:1:15
  |
1 | result: int = x + 1
  |               -^^^-
  |               |   |
  |               |   Has type `Literal[1]`
  |               Has type `Literal["hello"]`
  |
info: rule `unsupported-operator` is enabled by default

""")


def test_inject_stubs_offset():
    type_definitions = """\
from typing import Any

Messages = list[dict[str, Any]]

async def call_llm(prompt: str, messages: Messages) -> str | Messages:
    ...

prompt: str = ''
"""

    code = """\
async def agent(prompt: str, messages: Messages):
    while True:
        print(f'messages so far: {messages}')
        output = await call_llm(prompt, messages)
        if isinstance(output, str):
            return output
        messages.extend(output)

await agent(prompt, [])
"""
    ouros.Sandbox(
        code,
        inputs=['prompt'],
        external_functions=['call_llm'],
        script_name='agent.py',
        type_check=True,
        type_check_stubs=type_definitions,
    )

    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        ouros.Sandbox(
            code.replace('Messages', 'MXessages'),
            inputs=['prompt'],
            external_functions=['call_llm'],
            script_name='agent.py',
            type_check=True,
            type_check_stubs=type_definitions,
        )
    assert str(exc_info.value) == snapshot("""\
error[unresolved-reference]: Name `MXessages` used when not defined
 --> agent.py:1:40
  |
1 | async def agent(prompt: str, messages: MXessages):
  |                                        ^^^^^^^^^
2 |     while True:
3 |         print(f'messages so far: {messages}')
  |
info: rule `unresolved-reference` is enabled by default

""")

    code_call_func_wrong = 'await call_llm(prompt, 42)'

    with pytest.raises(ouros.SandboxTypingError) as exc_info:
        ouros.Sandbox(
            code_call_func_wrong,
            inputs=['prompt'],
            external_functions=['call_llm'],
            script_name='agent.py',
            type_check=True,
            type_check_stubs=type_definitions,
        )
    assert str(exc_info.value) == snapshot("""\
error[invalid-argument-type]: Argument to function `call_llm` is incorrect
 --> agent.py:1:24
  |
1 | await call_llm(prompt, 42)
  |                        ^^ Expected `list[dict[str, Any]]`, found `Literal[42]`
  |
info: Function defined here
 --> type_stubs.pyi:5:11
  |
3 | Messages = list[dict[str, Any]]
4 |
5 | async def call_llm(prompt: str, messages: Messages) -> str | Messages:
  |           ^^^^^^^^              ------------------ Parameter declared here
6 |     ...
  |
info: rule `invalid-argument-type` is enabled by default

""")



import asyncio

async def test_sleep():
    await asyncio.sleep(0)
    return 'done'

result = asyncio.run(test_sleep())
assert result == 'done', f'basic: {result}'

async def test_sleep_return():
    val = await asyncio.sleep(0)
    return val

result = asyncio.run(test_sleep_return())
assert result is None, f'return: {result}'

async def test_multiple():
    await asyncio.sleep(0)
    x = 1
    await asyncio.sleep(0)
    x += 1
    await asyncio.sleep(0)
    x += 1
    return x

result = asyncio.run(test_multiple())
assert result == 3, f'multiple: {result}'

print('ALL PASSED')



from typing import Any

import pytest
from inline_snapshot import snapshot

import ouros


def test_external_function_no_args():
    m = ouros.Sandbox('noop()', external_functions=['noop'])

    def noop(*args: Any, **kwargs: Any) -> str:
        assert args == snapshot(())
        assert kwargs == snapshot({})
        return 'called'

    assert m.run(external_functions={'noop': noop}) == snapshot('called')


def test_external_function_positional_args():
    m = ouros.Sandbox('func(1, 2, 3)', external_functions=['func'])

    def func(*args: Any, **kwargs: Any) -> str:
        assert args == snapshot((1, 2, 3))
        assert kwargs == snapshot({})
        return 'ok'

    assert m.run(external_functions={'func': func}) == snapshot('ok')


def test_external_function_kwargs_only():
    m = ouros.Sandbox('func(a=1, b="two")', external_functions=['func'])

    def func(*args: Any, **kwargs: Any) -> str:
        assert args == snapshot(())
        assert kwargs == snapshot({'a': 1, 'b': 'two'})
        return 'ok'

    assert m.run(external_functions={'func': func}) == snapshot('ok')


def test_external_function_mixed_args_kwargs():
    m = ouros.Sandbox('func(1, 2, x="hello", y=True)', external_functions=['func'])

    def func(*args: Any, **kwargs: Any) -> str:
        assert args == snapshot((1, 2))
        assert kwargs == snapshot({'x': 'hello', 'y': True})
        return 'ok'

    assert m.run(external_functions={'func': func}) == snapshot('ok')


def test_external_function_complex_types():
    m = ouros.Sandbox('func([1, 2], {"key": "value"})', external_functions=['func'])

    def func(*args: Any, **kwargs: Any) -> str:
        assert args == snapshot(([1, 2], {'key': 'value'}))
        assert kwargs == snapshot({})
        return 'ok'

    assert m.run(external_functions={'func': func}) == snapshot('ok')


def test_external_function_returns_none():
    m = ouros.Sandbox('do_nothing()', external_functions=['do_nothing'])

    def do_nothing(*args: Any, **kwargs: Any) -> None:
        assert args == snapshot(())
        assert kwargs == snapshot({})

    assert m.run(external_functions={'do_nothing': do_nothing}) is None


def test_external_function_returns_complex_type():
    m = ouros.Sandbox('get_data()', external_functions=['get_data'])

    def get_data(*args: Any, **kwargs: Any) -> dict[str, Any]:
        return {'a': [1, 2, 3], 'b': {'nested': True}}

    result = m.run(external_functions={'get_data': get_data})
    assert result == snapshot({'a': [1, 2, 3], 'b': {'nested': True}})


def test_multiple_external_functions():
    m = ouros.Sandbox('add(1, 2) + mul(3, 4)', external_functions=['add', 'mul'])

    def add(*args: Any, **kwargs: Any) -> int:
        assert args == snapshot((1, 2))
        assert kwargs == snapshot({})
        return args[0] + args[1]

    def mul(*args: Any, **kwargs: Any) -> int:
        assert args == snapshot((3, 4))
        assert kwargs == snapshot({})
        return args[0] * args[1]

    result = m.run(external_functions={'add': add, 'mul': mul})
    assert result == snapshot(15)  # 3 + 12


def test_external_function_called_multiple_times():
    m = ouros.Sandbox('counter() + counter() + counter()', external_functions=['counter'])

    call_count = 0

    def counter(*args: Any, **kwargs: Any) -> int:
        nonlocal call_count
        assert args == ()
        assert kwargs == {}
        call_count += 1
        return call_count

    result = m.run(external_functions={'counter': counter})
    assert result == snapshot(6)  # 1 + 2 + 3
    assert call_count == snapshot(3)


def test_external_function_with_input():
    m = ouros.Sandbox('process(x)', inputs=['x'], external_functions=['process'])

    def process(*args: Any, **kwargs: Any) -> int:
        assert args == snapshot((5,))
        assert kwargs == snapshot({})
        return args[0] * 10

    assert m.run(inputs={'x': 5}, external_functions={'process': process}) == snapshot(50)


def test_external_function_not_provided_raises():
    m = ouros.Sandbox('missing()', external_functions=['missing'])

    with pytest.raises(RuntimeError, match='no external_functions provided'):
        m.run()


def test_undeclared_function_raises_name_error():
    m = ouros.Sandbox('unknown_func()')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run()
    inner = exc_info.value.exception()
    assert type(inner) is NameError
    assert str(inner) == snapshot("name 'unknown_func' is not defined")


def test_external_function_raises_exception():
    """Test that exceptions from external functions propagate to the caller."""
    m = ouros.Sandbox('fail()', external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise ValueError('intentional error')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, ValueError)
    assert inner.args[0] == snapshot('intentional error')


def test_external_function_wrong_name_raises():
    """Test that calling a missing external function raises KeyError."""
    m = ouros.Sandbox('foo()', external_functions=['foo'])

    def bar(*args: Any, **kwargs: Any) -> int:
        return 1

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'bar': bar})
    inner = exc_info.value.exception()
    assert isinstance(inner, KeyError)
    # KeyError wraps the message in quotes
    assert inner.args[0] == snapshot('"External function \'foo\' not found"')


def test_external_function_exception_caught_by_try_except():
    """Test that exceptions from external functions can be caught by try/except."""
    code = """
try:
    fail()
except ValueError:
    caught = True
caught
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise ValueError('caught error')

    result = m.run(external_functions={'fail': fail})
    assert result == snapshot(True)


def test_external_function_exception_type_preserved():
    """Test that various exception types are correctly preserved."""
    m = ouros.Sandbox('fail()', external_functions=['fail'])

    def fail_type_error(*args: Any, **kwargs: Any) -> None:
        raise TypeError('type error message')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'fail': fail_type_error})
    inner = exc_info.value.exception()
    assert isinstance(inner, TypeError)
    assert inner.args[0] == snapshot('type error message')


@pytest.mark.parametrize(
    'exception_class,exception_name',
    [
        # ArithmeticError hierarchy
        (ZeroDivisionError, 'ZeroDivisionError'),
        (OverflowError, 'OverflowError'),
        (ArithmeticError, 'ArithmeticError'),
        # RuntimeError hierarchy
        (NotImplementedError, 'NotImplementedError'),
        (RecursionError, 'RecursionError'),
        (RuntimeError, 'RuntimeError'),
        # LookupError hierarchy
        (KeyError, 'KeyError'),
        (IndexError, 'IndexError'),
        (LookupError, 'LookupError'),
        # Other exceptions
        (ValueError, 'ValueError'),
        (TypeError, 'TypeError'),
        (AttributeError, 'AttributeError'),
        (NameError, 'NameError'),
        (AssertionError, 'AssertionError'),
    ],
)
def test_external_function_exception_hierarchy(exception_class: type[BaseException], exception_name: str):
    """Test that exception types in hierarchies are correctly preserved."""
    # Test that exception propagates with correct type
    m = ouros.Sandbox('fail()', external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise exception_class('test message')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, exception_class)


@pytest.mark.parametrize(
    'exception_class,parent_class,expected_result',
    [
        # ArithmeticError hierarchy
        (ZeroDivisionError, ArithmeticError, 'child'),
        (OverflowError, ArithmeticError, 'child'),
        # RuntimeError hierarchy
        (NotImplementedError, RuntimeError, 'child'),
        (RecursionError, RuntimeError, 'child'),
        # LookupError hierarchy
        (KeyError, LookupError, 'child'),
        (IndexError, LookupError, 'child'),
    ],
)
def test_external_function_exception_caught_by_parent(
    exception_class: type[BaseException], parent_class: type[BaseException], expected_result: str
):
    """Test that child exceptions can be caught by parent except handlers."""
    code = f"""
try:
    fail()
except {parent_class.__name__}:
    caught = 'parent'
except {exception_class.__name__}:
    caught = 'child'
caught
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise exception_class('test')

    # Child exception should be caught by parent handler (which comes first)
    result = m.run(external_functions={'fail': fail})
    assert result == 'parent'


@pytest.mark.parametrize(
    'exception_class,expected_result',
    [
        (ZeroDivisionError, 'ZeroDivisionError'),
        (OverflowError, 'OverflowError'),
        (NotImplementedError, 'NotImplementedError'),
        (RecursionError, 'RecursionError'),
        (KeyError, 'KeyError'),
        (IndexError, 'IndexError'),
    ],
)
def test_external_function_exception_caught_specifically(exception_class: type[BaseException], expected_result: str):
    """Test that child exceptions can be caught by their specific handler."""
    code = f"""
try:
    fail()
except {exception_class.__name__}:
    caught = '{expected_result}'
caught
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise exception_class('test')

    result = m.run(external_functions={'fail': fail})
    assert result == expected_result


def test_external_function_exception_in_expression():
    """Test exception from external function in an expression context."""
    m = ouros.Sandbox('1 + fail() + 2', external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> int:
        raise RuntimeError('mid-expression error')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, RuntimeError)
    assert inner.args[0] == snapshot('mid-expression error')


def test_external_function_exception_after_successful_call():
    """Test exception handling after a successful external call."""
    code = """
a = success()
b = fail()
a + b
"""
    m = ouros.Sandbox(code, external_functions=['success', 'fail'])

    def success(*args: Any, **kwargs: Any) -> int:
        return 10

    def fail(*args: Any, **kwargs: Any) -> int:
        raise ValueError('second call fails')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'success': success, 'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, ValueError)
    assert inner.args[0] == snapshot('second call fails')


def test_external_function_exception_with_finally():
    """Test that finally block runs when external function raises."""
    code = """
finally_ran = False
try:
    fail()
except ValueError:
    pass
finally:
    finally_ran = True
finally_ran
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail(*args: Any, **kwargs: Any) -> None:
        raise ValueError('error')

    result = m.run(external_functions={'fail': fail})
    assert result == snapshot(True)



# call-external
# External function calls in deep call stacks (function calling function).
# Tests that the outer function receives the return value correctly when
# the inner function makes an external call.


def depth1(n):
    return add_ints(n, 1)


def depth2(n):
    return depth1(n) + 10


result = depth2(5)
# depth2(5) should be: depth1(5) + 10 = 6 + 10 = 16

assert result == 16, f'ext call through 2 levels of functions {result=}'



# Test that accessing a variable before assignment at module level raises NameError
# (Unlike function scope, module level doesn't pre-scan for assignments)
print(x)
x = 1
"""
TRACEBACK:
Traceback (most recent call last):
  File "name_error__unbound_local_module.py", line 3, in <module>
    print(x)
          ~
NameError: name 'x' is not defined
"""



import html

# === escape ===
assert html.escape('') == '', 'escape_default'
assert html.escape(s='') == '', 'escape_required_keyword_form'
assert html.escape('hello') == 'hello', 'escape_combo_req_2'
assert html.escape('<b>hi</b>') == '&lt;b&gt;hi&lt;/b&gt;', 'escape_combo_req_3'
assert html.escape('', quote=True) == '', 'escape_opt_quote_1'
assert html.escape('', quote=False) == '', 'escape_opt_quote_2'

# === unescape ===
assert html.unescape('') == '', 'unescape_default'
assert html.unescape(s='') == '', 'unescape_required_keyword_form'
assert html.unescape('hello') == 'hello', 'unescape_combo_req_2'
assert html.unescape('<b>hi</b>') == '<b>hi</b>', 'unescape_combo_req_3'

# === roundtrip ===
assert html.unescape(html.escape('')) == '', 'roundtrip_escape_unescape_1'
assert html.unescape(html.escape('hello')) == 'hello', 'roundtrip_escape_unescape_2'
assert html.unescape(html.escape('<b>hi</b>')) == '<b>hi</b>', 'roundtrip_escape_unescape_3'
assert html.unescape(html.escape('"quotes"')) == '"quotes"', 'roundtrip_escape_unescape_4'



# call-external
# Test that accessing a non-existent attribute on a dataclass raises AttributeError
point = make_point()
point.z
"""
TRACEBACK:
Traceback (most recent call last):
  File "dataclass__get_missing_attr_error.py", line 4, in <module>
    point.z
AttributeError: 'Point' object has no attribute 'z'
"""



import hashlib


def hexdigest(result, length=None):
    if isinstance(result, str):
        return result
    if length is None:
        return result.hexdigest()
    return result.hexdigest(length)


def digest(result, length=None):
    if isinstance(result, bytes):
        return result
    if length is None:
        return result.digest()
    return result.digest(length)


# === new() dispatch ===
h = hashlib.new('sha256', b'hello')
assert hexdigest(h) == '2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824', 'new sha256'
assert h.digest_size == 32, 'sha256 digest_size attribute'
h = hashlib.new('sha512_224', b'hello')
assert hexdigest(h) == 'fe8509ed1fb7dcefc27e6ac1a80eddbec4cb3d2c6fe565244374061c', 'new sha512_224'
h = hashlib.new('sha512_256', b'hello')
assert hexdigest(h) == 'e30d87cfa2a75db545eac4d61baf970366a8357c7f72fa95b52d0accb698f13a', 'new sha512_256'

# These constructors are optional in CPython builds, but should be available in Ouros.
if hasattr(hashlib, 'sha512_224'):
    assert hexdigest(hashlib.sha512_224(b'hello')) == 'fe8509ed1fb7dcefc27e6ac1a80eddbec4cb3d2c6fe565244374061c', (
        'sha512_224 hello'
    )
if hasattr(hashlib, 'sha512_256'):
    assert (
        hexdigest(hashlib.sha512_256(b'hello')) == 'e30d87cfa2a75db545eac4d61baf970366a8357c7f72fa95b52d0accb698f13a'
    ), 'sha512_256 hello'

# === sha3 family ===
assert hexdigest(hashlib.sha3_224(b'hello')) == 'b87f88c72702fff1748e58b87e9141a42c0dbedc29a78cb0d4a5cd81', (
    'sha3_224 hello'
)
assert hexdigest(hashlib.sha3_256(b'hello')) == '3338be694f50c5f338814986cdf0686453a888b84f424d792af4b9202398f392', (
    'sha3_256 hello'
)
assert hexdigest(hashlib.sha3_384(b'hello')) == (
    '720aea11019ef06440fbf05d87aa24680a2153df3907b23631e7177ce620fa1330ff07c0fddee54699a4c3ee0ee9d887'
), 'sha3_384 hello'
assert hexdigest(hashlib.sha3_512(b'hello')) == (
    '75d527c368f2efe848ecf6b073a36767800805e9eef2b1857d5f984f036eb6df'
    '891d75f72d9b154518c1cd58835286d1da9a38deba3de98b5a53e5ed78a84976'
), 'sha3_512 hello'

# === blake2 family ===
assert hexdigest(hashlib.blake2b(b'hello')) == (
    'e4cfa39a3d37be31c59609e807970799caa68a19bfaa15135f165085e01d41a6'
    '5ba1e1b146aeb6bd0092b49eac214c103ccfa3a365954bbbe52f74a2b3620c94'
), 'blake2b hello'
assert hexdigest(hashlib.blake2s(b'hello')) == '19213bacc58dee6dbde3ceb9a47cbb330b3d86f8cca8997eb00be456f140ca25', (
    'blake2s hello'
)
assert hexdigest(hashlib.new('blake2b', b'hello')) == hexdigest(hashlib.blake2b(b'hello')), 'new blake2b'
assert hexdigest(hashlib.new('blake2s', b'hello')) == hexdigest(hashlib.blake2s(b'hello')), 'new blake2s'

# === shake family ===
assert hexdigest(hashlib.shake_128(b'hello'), 16) == '8eb4b6a932f280335ee1a279f8c208a3', 'shake_128 hello hex'
assert hexdigest(hashlib.shake_256(b'hello'), 16) == '1234075ae4a1e77316cf2d8000974581', 'shake_256 hello hex'
assert len(digest(hashlib.shake_128(b'hello'), 16)) == 16, 'shake_128 digest length'

# === algorithms_* ===
assert isinstance(hashlib.algorithms_available, set), 'algorithms_available should be set'
assert isinstance(hashlib.algorithms_guaranteed, set), 'algorithms_guaranteed should be set'
assert 'sha256' in hashlib.algorithms_available, 'sha256 should be available'
assert 'sha512_224' in hashlib.algorithms_available, 'sha512_224 should be available'
assert 'sha512_256' in hashlib.algorithms_available, 'sha512_256 should be available'
assert 'blake2b' in hashlib.algorithms_available, 'blake2b should be available'
assert 'blake2s' in hashlib.algorithms_available, 'blake2s should be available'
assert 'sha256' in hashlib.algorithms_guaranteed, 'sha256 should be guaranteed'

# === pbkdf2_hmac ===
sha256_key = hashlib.pbkdf2_hmac('sha256', b'password', b'salt', 1000, 32)
assert sha256_key.hex() == '632c2812e46d4604102ba7618e9d6d7d2f8128f6266b4a03264d2a0460b7dcb3', 'pbkdf2_hmac sha256'

sha1_key = hashlib.pbkdf2_hmac('sha1', b'password', b'salt', 1000)
assert sha1_key.hex() == '6e88be8bad7eae9d9e10aa061224034fed48d03f', 'pbkdf2_hmac sha1 default dklen'

# === scrypt (stub) ===
try:
    scrypt_bytes = hashlib.scrypt(b'password', salt=b'salt', n=2**14, r=8, p=1)
    assert isinstance(scrypt_bytes, bytes), 'scrypt should return bytes when implemented'
except NotImplementedError as e:
    assert str(e) == 'hashlib.scrypt() is not implemented', f'unexpected scrypt message: {e}'



# call-external
# === External calls in if/else expressions ===

# Ternary expression with ext call in condition
result = 'yes' if add_ints(1, 1) == 2 else 'no'
assert result == 'yes', 'ext call in ternary condition true'

result2 = 'yes' if add_ints(1, 1) == 3 else 'no'
assert result2 == 'no', 'ext call in ternary condition false'

# Ext call in true branch
val = add_ints(10, 20) if True else 0
assert val == 30, 'ext call in ternary true branch'

# Ext call in false branch
val2 = 0 if False else add_ints(5, 5)
assert val2 == 10, 'ext call in ternary false branch'

# Ext calls in both branches
val3 = add_ints(1, 2) if True else add_ints(3, 4)
assert val3 == 3, 'ext call in both branches takes true'

val4 = add_ints(1, 2) if False else add_ints(3, 4)
assert val4 == 7, 'ext call in both branches takes false'

# === If statements with external calls ===

# Ext call in if condition
x = 0
if add_ints(1, 1) == 2:
    x = 100
assert x == 100, 'ext call in if statement condition true'

y = 0
if add_ints(1, 1) == 3:
    y = 100
assert y == 0, 'ext call in if statement condition false'

# Ext call in if body
z = 0
if True:
    z = add_ints(50, 50)
assert z == 100, 'ext call in if body'

# Ext call in else body
w = 0
if False:
    w = 1
else:
    w = add_ints(25, 75)
assert w == 100, 'ext call in else body'

# Nested ext calls in condition
nested = 0
if add_ints(add_ints(1, 2), add_ints(3, 4)) == 10:
    nested = 1
assert nested == 1, 'nested ext calls in if condition'

# Chained conditions with ext calls
result3 = 0
if add_ints(1, 1) == 2 and add_ints(2, 2) == 4:
    result3 = 1
assert result3 == 1, 'multiple ext calls in and condition'

result4 = 0
if add_ints(1, 1) == 3 or add_ints(2, 2) == 4:
    result4 = 1
assert result4 == 1, 'multiple ext calls in or condition'

# Comparison with ext call results
cmp = add_ints(10, 5) > add_ints(5, 5)
assert cmp == True, 'comparing two ext call results'

# === Nested if statements with external calls ===

# Nested if with ext calls in both conditions (both true)
result = 'none'
if return_value(1) == 1:
    if return_value(2) == 2:
        result = 'inner'
    else:
        result = 'outer_only'
else:
    result = 'failed'
assert result == 'inner', 'nested if both conditions true'

# Nested if - outer true, inner false
result2 = 'none'
if return_value(1) == 1:
    if return_value(2) == 999:
        result2 = 'inner'
    else:
        result2 = 'outer_only'
else:
    result2 = 'failed'
assert result2 == 'outer_only', 'nested if inner false'

# Nested if - outer false
result3 = 'none'
if return_value(1) == 999:
    if return_value(2) == 2:
        result3 = 'inner'
    else:
        result3 = 'outer_only'
else:
    result3 = 'xxx'
assert result3 == 'xxx', 'nested if outer false'

# Triple nested if - all true
result4 = 0
if return_value(1) == 1:
    if return_value(2) == 2:
        if return_value(3) == 3:
            result4 = 123
assert result4 == 123, 'triple nested if all true'

# If condition with multiple ext calls (addition)
result5 = 0
if add_ints(1, 2) + add_ints(3, 4) == 10:
    result5 = 1
assert result5 == 1, 'if condition with multiple ext calls'

# For loop inside if with external condition
total = 0
if return_value(1) == 1:
    for i in range(3):
        total = add_ints(total, return_value(i))
assert total == 3, 'for loop inside if with ext condition'

# For loop inside if - condition false
total2 = 0
if return_value(1) == 999:
    for i in range(3):
        total2 = add_ints(total2, i)
assert total2 == 0, 'for loop inside if condition false'



# Tests reference counting correctness for **kwargs unpacking


def receive_kwargs(a, b, c):
    return a


# === Heap-allocated values in unpacked dict ===
# All heap objects must be directly referenced by variables for strict matching
list_a = [1, 2, 3]
list_c = [4, 5]
kwargs_dict = {'a': list_a, 'b': 'hello', 'c': list_c}
result = receive_kwargs(**kwargs_dict)
assert result == [1, 2, 3], 'received list via **kwargs'
assert result is list_a, 'should be same object'

# Second call to verify dict reuse works
result2 = receive_kwargs(**kwargs_dict)
assert result2 is list_a, 'second call returns same object'

# list_a: 5 refs (list_a var, kwargs_dict['a'], result, result2, final expr)
# list_c: 2 refs (list_c var, kwargs_dict['c'])
# kwargs_dict: 1 ref
# result: 5 refs (same object as list_a)
# result2: 5 refs (same object as list_a)
result2
# ref-counts={'list_a': 5, 'list_c': 2, 'kwargs_dict': 1, 'result': 5, 'result2': 5}



# Basic class definition and instantiation


class Dog:
    species = 'Canis familiaris'

    def __init__(self, name, age):
        self.name = name
        self.age = age

    def bark(self):
        return f'{self.name} says woof'

    def get_age(self):
        return self.age


d = Dog('Rex', 5)
assert d.name == 'Rex', f"expected 'Rex', got {d.name}"
assert d.age == 5, f'expected 5, got {d.age}'
assert d.bark() == 'Rex says woof', f'unexpected bark: {d.bark()}'
assert Dog.species == 'Canis familiaris', 'class attr access failed'
assert d.species == 'Canis familiaris', 'instance class attr access failed'

# Multiple instances are independent
d2 = Dog('Buddy', 3)
assert d2.name == 'Buddy'
assert d.name == 'Rex'

# type() returns the class
assert type(d) is Dog, f'type mismatch: {type(d)}'

# Instance attribute shadows class attribute
d.species = 'wolf'
assert d.species == 'wolf', 'instance shadow failed'
assert Dog.species == 'Canis familiaris', 'class attr mutated'


# Class with no __init__
class Empty:
    pass


e = Empty()
assert type(e) is Empty


# Class attribute mutation via method
class Counter:
    count = 0

    def __init__(self):
        c = Counter.count
        Counter.count = c + 1


Counter()
Counter()
assert Counter.count == 2, f'expected 2, got {Counter.count}'


# Method returning self (chaining)
class Builder:
    def __init__(self):
        self.parts = []

    def add(self, part):
        self.parts.append(part)
        return self


b = Builder()
b.add('a').add('b').add('c')
assert b.parts == ['a', 'b', 'c'], f'chaining failed: {b.parts}'


# Circular references don't crash
class Node:
    def __init__(self):
        self.next = None


a = Node()
b = Node()
a.next = b
b.next = a  # Circular reference — should not crash



from urllib.parse import urlparse

r = urlparse('https://api.example.com:8443/v2/users?page=3')
assert r.hostname == 'api.example.com', f'hostname: {r.hostname}'
assert r.port == 8443, f'port: {r.port}'

r = urlparse('https://example.com/path')
assert r.hostname == 'example.com'
assert r.port is None

r = urlparse('https://user:pass@host.com:443/path')
assert r.hostname == 'host.com'
assert r.port == 443
assert r.username == 'user'
assert r.password == 'pass'

r = urlparse('https://example.com/path')
assert r.username is None
assert r.password is None

r = urlparse('https://EXAMPLE.COM/path')
assert r.hostname == 'example.com'

r = urlparse('/relative/path')
assert r.hostname is None
assert r.port is None

print('ALL PASSED')



# run-async
import asyncio


# === Basic gather ===
async def task1():
    return 1


async def task2():
    return 2


result = await asyncio.gather(task1(), task2())  # pyright: ignore
assert result == [1, 2], 'gather should return results as a list'


# === Result ordering ===
# Results should be in argument order, not completion order
async def slow():
    return 'slow'


async def fast():
    return 'fast'


result = await asyncio.gather(slow(), fast())  # pyright: ignore
assert result == ['slow', 'fast'], 'gather should preserve argument order'

# === Empty gather ===
result = await asyncio.gather()  # pyright: ignore
assert result == [], 'empty gather should return empty list'


# === Single coroutine ===
async def single():
    return 42


result = await asyncio.gather(single())  # pyright: ignore
assert result == [42], 'gather with single coroutine should return list with one element'

# === repr of gather function ===
r = repr(asyncio.gather)
assert r.startswith('<function gather at 0x'), f'repr should start with: {r}'

# === TypeError for non-awaitable argument ===
try:
    await asyncio.gather(123)  # pyright: ignore
    assert False, 'should have raised TypeError'
except TypeError as e:
    assert str(e) == 'An asyncio.Future, a coroutine or an awaitable is required'


# === *args unpacking with gather ===
async def a():
    return 'a'


async def b():
    return 'b'


async def c():
    return 'c'


# Unpack a list of coroutines
coros = [a(), b(), c()]
result = await asyncio.gather(*coros)  # pyright: ignore
assert result == ['a', 'b', 'c'], f'gather with *args unpacking: {result}'

# Unpack with mixed args
result = await asyncio.gather(a(), *[b(), c()])  # pyright: ignore
assert result == ['a', 'b', 'c'], f'gather with mixed args and *unpacking: {result}'

# Unpack empty list
result = await asyncio.gather(*[])  # pyright: ignore
assert result == [], f'gather with empty *args: {result}'

# Unpack tuple
coro_tuple = (a(), b())
result = await asyncio.gather(*coro_tuple)  # pyright: ignore
assert result == ['a', 'b'], f'gather with *tuple unpacking: {result}'



# call-external
# === External calls in elif conditions ===

# External call in elif condition - true
result1 = 0
if add_ints(1, 1) == 3:
    result1 = 1
elif add_ints(2, 2) == 4:
    result1 = 2
else:
    result1 = 3
assert result1 == 2, f'elif condition with ext call should be evaluated, {result1=}'

# External call in elif condition - falls through to else
result2 = 0
if add_ints(1, 1) == 3:
    result2 = 1
elif add_ints(2, 2) == 5:
    result2 = 2
else:
    result2 = 3
assert result2 == 3, 'else taken when all ext call conditions are false'

# Multiple elif with external calls
result3 = 0
if add_ints(1, 1) == 10:
    result3 = 1
elif add_ints(2, 2) == 10:
    result3 = 2
elif add_ints(3, 3) == 6:
    result3 = 3
elif add_ints(4, 4) == 10:
    result3 = 4
else:
    result3 = 5
assert result3 == 3, 'third elif with ext call should match'

# === External calls in elif bodies ===

# Ext call in elif body
val1 = 0
if False:
    val1 = 1
elif True:
    val1 = add_ints(10, 20)
else:
    val1 = 3
assert val1 == 30, 'ext call in elif body'

# Ext call in else body after elif chain
val2 = 0
if False:
    val2 = 1
elif False:
    val2 = 2
else:
    val2 = add_ints(15, 25)
assert val2 == 40, 'ext call in else body after elif'

# Multiple ext calls in elif body
val3 = 0
if False:
    val3 = 1
elif True:
    a = add_ints(5, 5)
    b = add_ints(10, 10)
    val3 = add_ints(a, b)
else:
    val3 = 3
assert val3 == 30, 'multiple ext calls in elif body'

# === Nested ext calls ===

# Nested ext calls in elif condition
result4 = 0
if False:
    result4 = 1
elif add_ints(add_ints(1, 2), add_ints(3, 4)) == 10:
    result4 = 2
else:
    result4 = 3
assert result4 == 2, 'nested ext calls in elif condition'

# === Short-circuit with ext calls ===

# Ext call should not be evaluated if earlier condition is true
call_count = 0


def counting_add(a, b):
    global call_count
    call_count = call_count + 1
    return a + b


# This uses a regular function, not ext call, to verify short-circuit
# but the ext calls in bodies still test suspension
x = 0
if True:
    x = add_ints(1, 1)
elif False:
    x = add_ints(2, 2)
assert x == 2, 'if body ext call executed, elif skipped'

# === Ext call in both condition and body ===

result5 = 0
if add_ints(1, 1) == 3:
    result5 = add_ints(100, 100)
elif add_ints(2, 2) == 4:
    result5 = add_ints(50, 50)
else:
    result5 = add_ints(25, 25)
assert result5 == 100, 'ext call in both elif condition and body'

# === Ext call in if body when condition is true ===

if_body_result = 0
if add_ints(5, 5) == 10:
    if_body_result = add_ints(100, 200)
elif add_ints(1, 1) == 2:
    if_body_result = add_ints(10, 20)
else:
    if_body_result = add_ints(1, 2)
assert if_body_result == 300, 'ext call in if body when if condition is true'

# === Ext calls returning values used as conditions ===

# return_value returns its argument, so we can use it to test boolean coercion
cond_result = 0
if return_value(0):
    cond_result = 1
elif return_value(1):
    cond_result = 2
else:
    cond_result = 3
assert cond_result == 2, 'ext call return value used as boolean condition'

# === Ext calls with string concatenation ===

str_result = ''
if add_ints(1, 1) == 3:
    str_result = concat_strings('a', 'b')
elif add_ints(2, 2) == 4:
    str_result = concat_strings('hello', ' world')
else:
    str_result = concat_strings('x', 'y')
assert str_result == 'hello world', 'ext call with string result in elif body'

# === Multiple conditions with ext calls in same expression ===

multi_cond = 0
if add_ints(1, 1) > 5:
    multi_cond = 1
elif add_ints(2, 2) < add_ints(3, 3):
    multi_cond = 2
else:
    multi_cond = 3
assert multi_cond == 2, 'comparison between two ext call results in elif condition'

# === Ext call in all three branches ===

all_branches = 0
val = add_ints(5, 5)
if val < 5:
    all_branches = add_ints(1, 0)
elif val < 15:
    all_branches = add_ints(2, 0)
else:
    all_branches = add_ints(3, 0)
assert all_branches == 2, 'ext call in elif body based on earlier ext call result'



# Test that accessing an undefined global name raises NameError
unknown_func()
"""
TRACEBACK:
Traceback (most recent call last):
  File "name_error__undefined_global.py", line 2, in <module>
    unknown_func()
    ~~~~~~~~~~~~
NameError: name 'unknown_func' is not defined
"""



# call-external
# BUG: Side effects in arguments are duplicated when external call suspends
#
# When an external call is in one argument position and there's a side effect
# in another argument position, the side effect may be executed multiple times
# because argument evaluation is repeated during resumption.

call_count = 0


def side_effect(val):
    global call_count
    call_count += 1
    return val


# Side effect before external call in args
# Expected: side_effect runs once, result is 10 + 3 = 13
call_count = 0
result = add_ints(side_effect(10), add_ints(1, 2))
assert result == 13, 'ext call after side effect'
assert call_count == 1, 'side effect should happen only once (before ext)'



# === Basic **kwargs unpacking ===
def greet(name, greeting):
    return f'{greeting}, {name}!'


opts = {'greeting': 'Hi'}
assert greet('Alice', **opts) == 'Hi, Alice!', 'basic **kwargs unpacking'

# === Dict literal unpacking ===
assert greet('Charlie', **{'greeting': 'Hey'}) == 'Hey, Charlie!', 'dict literal unpacking'


# === Multiple kwargs in unpacked dict ===
def format_msg(msg, prefix, suffix):
    return f'{prefix}{msg}{suffix}'


assert format_msg('test', **{'prefix': '[', 'suffix': ']'}) == '[test]', 'multiple kwargs unpacking'

# === Combining regular kwargs with **kwargs ===
assert format_msg('hello', prefix='> ', **{'suffix': '!'}) == '> hello!', 'regular kwargs with **kwargs'


# === **kwargs with positional args ===
def add_all(a, b, c):
    return a + b + c


assert add_all(1, 2, **{'c': 3}) == 6, '**kwargs with positional args'
assert add_all(1, **{'b': 2, 'c': 3}) == 6, '**kwargs providing multiple args'

# === Variable dict unpacking ===
settings = {'prefix': '>>> ', 'suffix': ' <<<'}
assert format_msg('output', **settings) == '>>> output <<<', 'variable dict unpacking'


# === Unpacking with keyword-only args ===
def kwonly_func(a, *, b, c):
    return a + b + c


assert kwonly_func(1, **{'b': 2, 'c': 3}) == 6, '**kwargs with keyword-only args'


# === Empty dict unpacking with all args provided ===
def simple(x, y):
    return x + y


assert simple(1, 2, **{}) == 3, 'empty dict unpacking'


# === All kwargs from unpacking ===
def all_kwargs(a, b, c):
    return a * 100 + b * 10 + c


assert all_kwargs(**{'a': 1, 'b': 2, 'c': 3}) == 123, 'all args from **kwargs'
assert all_kwargs(**{'c': 7, 'a': 4, 'b': 5}) == 457, 'all args from **kwargs different order'


# === Dynamic **kwargs keys ===
def kwonly_echo(*, keyword):
    return keyword


key_name = 'k' + 'e' + 'y' + 'w' + 'o' + 'r' + 'd'
assert kwonly_echo(**{key_name: 'dynamic'}) == 'dynamic', 'runtime string key matches kw-only param'


# ============================================================
# *args unpacking tests (function calls)
# ============================================================


# === *args with zero args ===
def no_args():
    return 'ok'


assert no_args(*[]) == 'ok', '*args with empty list'
assert no_args(*()) == 'ok', '*args with empty tuple'


# === *args with one arg ===
def one_arg(x):
    return x * 2


assert one_arg(*[5]) == 10, '*args with one item list'
assert one_arg(*(7,)) == 14, '*args with one item tuple'


# === *args with two args ===
def two_args(a, b):
    return a + b


assert two_args(*[1, 2]) == 3, '*args with two item list'
assert two_args(*(3, 4)) == 7, '*args with two item tuple'


# === *args with three+ args ===
def many_args(a, b, c, d):
    return a + b + c + d


assert many_args(*[1, 2, 3, 4]) == 10, '*args with four items'
assert many_args(*(10, 20, 30, 40)) == 100, '*args with tuple four items'


# === Mixed positional and *args ===
assert two_args(1, *[2]) == 3, 'pos + *args'
assert many_args(1, 2, *[3, 4]) == 10, 'two pos + *args'


# === *args with heap-allocated values ===
def list_arg(lst):
    return len(lst)


my_list = [1, 2, 3]
assert list_arg(*[my_list]) == 3, '*args with list value'


# ============================================================
# Combined *args and **kwargs (function calls)
# ============================================================


# === *args and **kwargs together ===
def mixed_func(a, b, c):
    return f'{a}-{b}-{c}'


assert mixed_func(*[1], **{'b': 2, 'c': 3}) == '1-2-3', '*args and **kwargs'
assert mixed_func(*[1, 2], **{'c': 3}) == '1-2-3', 'two *args and **kwargs'


# === *args tuple with **kwargs ===
args_tuple = (10, 20)
kwargs_dict = {'c': 30}
assert many_args(*args_tuple, **kwargs_dict, d=40) == 100, '*args tuple + **kwargs + regular kwarg'


# === Empty *args with **kwargs ===
assert mixed_func(*[], **{'a': 'x', 'b': 'y', 'c': 'z'}) == 'x-y-z', 'empty *args with **kwargs'


# === *args with empty **kwargs ===
assert two_args(*[5, 6], **{}) == 11, '*args with empty **kwargs'


# === All combinations: pos, *args, kwargs, **kwargs ===
def full_func(a, b, c, d):
    return a * 1000 + b * 100 + c * 10 + d


assert full_func(1, *[2], c=3, **{'d': 4}) == 1234, 'pos + *args + kwarg + **kwargs'


# === *args with heap values and **kwargs ===
def heap_func(lst, dct):
    return len(lst) + len(dct)


list_val = [1, 2, 3]
dict_val = {'a': 1}
assert heap_func(*[list_val], **{'dct': dict_val}) == 4, '*args and **kwargs with heap values'


# === Both *args and **kwargs empty ===
assert no_args(*[], **{}) == 'ok', 'empty *args and empty **kwargs'



import base64

# === standard_b64encode/standard_b64decode ===
assert base64.standard_b64encode(b'hello') == base64.b64encode(b'hello'), 'standard_b64encode alias'
assert base64.standard_b64decode(b'aGVsbG8=') == base64.b64decode(b'aGVsbG8='), 'standard_b64decode alias'
assert base64.b64encode(bytes([0, 1, 2])) == b'AAEC', 'b64encode bytes(list)'
assert base64.b64decode(base64.b64encode(bytes(range(4)))) == bytes(range(4)), 'b64 roundtrip bytes(range)'

# === b32hexencode/b32hexdecode ===
assert base64.b32hexencode(b'') == b'', 'b32hex encode empty'
assert base64.b32hexencode(b'hello') == b'D1IMOR3F', 'b32hex encode hello'
assert base64.b32hexencode(b'foobar') == b'CPNMUOJ1E8======', 'b32hex encode foobar'
assert base64.b32hexdecode(b'D1IMOR3F') == b'hello', 'b32hex decode hello'
assert base64.b32hexdecode(b'CPNMUOJ1E8======') == b'foobar', 'b32hex decode foobar'
for text in [b'1234567', b'ouros']:
    assert base64.b32hexdecode(base64.b32hexencode(text)) == text, 'b32hex roundtrip ' + text.decode()

# === encodebytes/decodebytes ===
assert base64.encodebytes(b'') == b'', 'encodebytes empty'
expected_57 = b'YWFh' * 19 + b'\n'
assert base64.encodebytes(b'a' * 57) == expected_57, 'encodebytes wraps at 76'
expected_58 = b'YWFh' * 19 + b'\n' + b'YQ==\n'
assert base64.encodebytes(b'a' * 58) == expected_58, 'encodebytes wraps with remainder'
assert base64.decodebytes(b'Y Q==\n') == b'a', 'decodebytes strips whitespace'

# === a85encode/a85decode ===
assert base64.a85encode(b'') == b'', 'a85 encode empty'
assert base64.a85encode(b'1234') == b'0etOA', 'a85 encode 1234'
assert base64.a85encode(b'123') == b'0etN', 'a85 encode 123'
assert base64.a85encode(b'12') == b'0er', 'a85 encode 12'
assert base64.a85encode(b'1') == b'0`', 'a85 encode 1'
assert base64.a85encode(b'\x00\x00\x00\x00') == b'z', 'a85 encode zero quad'
assert base64.a85encode(b'hello') == b'BOu!rDZ', 'a85 encode hello'
assert base64.a85decode(b'BOu!rDZ') == b'hello', 'a85 decode hello'
assert base64.a85decode(b'BOu!r D Z') == b'hello', 'a85 decode ignores whitespace'
assert base64.a85decode(b'z') == b'\x00\x00\x00\x00', 'a85 decode zero quad'
for text in [b'ouros', b'hello world', b'12345']:
    assert base64.a85decode(base64.a85encode(text)) == text, 'a85 roundtrip ' + text.decode()

# === z85encode/z85decode ===
assert base64.z85encode(b'') == b'', 'z85 encode empty'
assert base64.z85encode(b'1234') == b'f!$Kw', 'z85 encode 1234'
assert base64.z85encode(b'123') == b'f!$J', 'z85 encode 123'
assert base64.z85encode(b'1') == b'f-', 'z85 encode 1'
assert base64.z85encode(b'hello') == b'xK#0@zV', 'z85 encode hello'
assert base64.z85encode(b'\x00\x00\x00\x00') == b'00000', 'z85 encode zero quad'
assert base64.z85decode(b'f!$Kw') == b'1234', 'z85 decode 1234'
assert base64.z85decode(b'f!$J') == b'123', 'z85 decode 123'
assert base64.z85decode(b'f-') == b'1', 'z85 decode 1'
assert base64.z85decode(b'xK#0@zV') == b'hello', 'z85 decode hello'
for text in [b'hello!!!!', b'1234567']:
    assert base64.z85decode(base64.z85encode(text)) == text, 'z85 roundtrip ' + text.decode()


# === encode/decode stubs ===
class DummyEncode:
    def read(self, *args, **kwargs):
        raise NotImplementedError('base64.encode requires file objects')

    def write(self, *args, **kwargs):
        raise NotImplementedError('base64.encode requires file objects')


class DummyDecode:
    def readline(self, *args, **kwargs):
        raise NotImplementedError('base64.decode requires file objects')

    def read(self, *args, **kwargs):
        raise NotImplementedError('base64.decode requires file objects')

    def write(self, *args, **kwargs):
        raise NotImplementedError('base64.decode requires file objects')


try:
    base64.encode(DummyEncode(), DummyEncode())
    assert False, 'encode should raise NotImplementedError'
except NotImplementedError as exc:
    assert str(exc) == 'base64.encode requires file objects', 'encode error message'

try:
    base64.decode(DummyDecode(), DummyDecode())
    assert False, 'decode should raise NotImplementedError'
except NotImplementedError as exc:
    assert str(exc) == 'base64.decode requires file objects', 'decode error message'

# === String input to decode functions ===
# CPython's base64 decode functions accept both bytes and ASCII strings
assert base64.b64decode('aGVsbG8=') == b'hello', 'b64decode accepts string input'
assert base64.b32decode('JBSWY3DPEBLW64TMMQ======') == b'Hello World', 'b32decode accepts string input'
assert base64.b16decode('48656C6C6F') == b'Hello', 'b16decode accepts string input'

# === casefold parameter for b32decode ===
# When casefold=True, lowercase characters are accepted
assert base64.b32decode('jbswy3dpeblw64tmmq======', casefold=True) == b'Hello World', 'b32decode casefold=True'

# === casefold parameter for b16decode ===
# When casefold=True, lowercase hex characters are accepted
assert base64.b16decode('48656c6c6f', casefold=True) == b'Hello', 'b16decode casefold=True'

# === b32hexdecode with casefold ===
assert base64.b32hexdecode('d1imor3f', casefold=True) == b'hello', 'b32hexdecode casefold=True'



'hello'.rsplit('')
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__rsplit_empty_sep.py", line 1, in <module>
    'hello'.rsplit('')
    ~~~~~~~~~~~~~~~~~~
ValueError: empty separator
"""
